/*
 * Copyright (C) 2005-2012 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file audiod_command.c Commands for para_audiod. */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "audiod.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "filter.h"
#include "grab_client.h"
#include "error.h"
#include "audiod.h"
#include "net.h"
#include "daemon.h"
#include "string.h"
#include "write.h"
#include "fd.h"
#include "audiod_command_list.h"

extern struct sched sched;
extern char *stat_item_values[NUM_STAT_ITEMS];


struct audiod_command audiod_cmds[] = {DEFINE_AUDIOD_CMD_ARRAY};

/** Iterate over the array of all audiod commands. */
#define FOR_EACH_COMMAND(c) for (c = 0; audiod_cmds[c].name; c++)

/** The maximal number of simultaneous connections. */
#define MAX_STAT_CLIENTS 50

/** Flags used for the stat command of para_audiod. */
enum stat_client_flags {
	/** Enable parser-friendly output. */
	SCF_PARSER_FRIENDLY = 1,
};

/**
 * Describes a status client of para_audiod.
 *
 * There's one such structure per audiod client that sent the 'stat' command.
 *
 * A status client is identified by its file descriptor.  para_audiod
 * keeps a list of connected status clients.
 */
struct stat_client {
	/** The stat client's file descriptor. */
	int fd;
	/** Bitmask of those status items the client is interested in. */
	uint64_t item_mask;
	/** See \ref stat_client flags. s*/
	unsigned flags;
	/** Its entry in the list of stat clients. */
	struct list_head node;
};

static INITIALIZED_LIST_HEAD(client_list);
static int num_clients;

/** The list of all status items used by para_{server,audiod,gui}. */
const char *status_item_list[] = {STATUS_ITEM_ARRAY};

static void dump_stat_client_list(void)
{
	struct stat_client *sc;

	list_for_each_entry(sc, &client_list, node)
		PARA_INFO_LOG("stat client on fd %d\n", sc->fd);
}
/**
 * Add a status client to the list.
 *
 * \param fd The file descriptor of the client.
 * \param mask Bitfield of status items for this client.
 * \param parser_friendly Enable parser-friendly output mode.
 *
 * Only those status items having the bit set in \a mask will be
 * sent to the client.
 *
 * \return Positive value on success, or -E_TOO_MANY_CLIENTS if
 * the number of connected clients exceeds #MAX_STAT_CLIENTS.
 */
static int stat_client_add(int fd, uint64_t mask, int parser_friendly)
{
	struct stat_client *new_client;

	if (num_clients >= MAX_STAT_CLIENTS) {
		PARA_ERROR_LOG("maximal number of stat clients (%d) exceeded\n",
			MAX_STAT_CLIENTS);
		return -E_TOO_MANY_CLIENTS;
	}
	PARA_INFO_LOG("adding client on fd %d\n", fd);
	new_client = para_calloc(sizeof(struct stat_client));
	new_client->fd = fd;
	new_client->item_mask = mask;
	if (parser_friendly)
		new_client->flags = SCF_PARSER_FRIENDLY;
	para_list_add(&new_client->node, &client_list);
	dump_stat_client_list();
	num_clients++;
	return 1;
}
/**
 * Write a message to all connected status clients.
 *
 * \param item_num The number of the status item of \a msg.
 *
 * On write errors, remove the status client from the client list and close its
 * file descriptor.
 */
void stat_client_write_item(int item_num)
{
	struct stat_client *sc, *tmp;
	struct para_buffer pb = {.flags = 0};
	struct para_buffer pfpb = {.flags = PBF_SIZE_PREFIX};
	const uint64_t one = 1;
	char *msg = stat_item_values[item_num];
	struct para_buffer *b;

	list_for_each_entry_safe(sc, tmp, &client_list, node) {
		int fd = sc->fd, ret;

		if (!((one << item_num) & sc->item_mask))
			continue;
		b = (sc->flags & SCF_PARSER_FRIENDLY)? &pfpb : &pb;
		if (!b->buf)
			(void)WRITE_STATUS_ITEM(b, item_num, "%s\n",
				msg? msg : "");
		ret = write(fd, b->buf, b->offset);
		if (ret == b->offset)
			continue;
		/* write error or short write */
		close(fd);
		num_clients--;
		PARA_INFO_LOG("deleting client on fd %d\n", fd);
		list_del(&sc->node);
		free(sc);
		dump_stat_client_list();
	}
	free(pb.buf);
	free(pfpb.buf);
}

/**
 * Check if string is a known status item.
 *
 * \param item Buffer containing the text to check.
 *
 * \return If \a item is a valid status item, the number of that status item is
 * returned. Otherwise, this function returns \p -E_UNKNOWN_STAT_ITEM.
 */
static int stat_item_valid(const char *item)
{
	int i;
	if (!item || !*item) {
		PARA_ERROR_LOG("%s\n", "no item");
		return -E_UNKNOWN_STAT_ITEM;
	}
	FOR_EACH_STATUS_ITEM(i)
		if (!strcmp(status_item_list[i], item))
			return i;
	PARA_ERROR_LOG("invalid stat item: %s\n", item);
	return -E_UNKNOWN_STAT_ITEM;
}

static int client_write(int fd, const char *buf)
{
	size_t len;

	if (!buf)
		return 0;
	len = strlen(buf);
	return write(fd, buf, len) != len? -E_CLIENT_WRITE: 1;
}

__malloc static char *audiod_status_string(void)
{
	const char *status = (audiod_status == AUDIOD_ON)?
		"on" : (audiod_status == AUDIOD_OFF)? "off": "sb";
	return para_strdup(status);
}

static int get_play_time_slot_num(void)
{
	int i, oldest_slot = -1;
	struct timeval oldest_wstime = {0, 0};

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		struct timeval wstime;
		if (!s->wns || !s->wns[0].btrn)
			continue;
		btr_get_node_start(s->wns[0].btrn, &wstime);
		if (oldest_slot >= 0 && tv_diff(&wstime, &oldest_wstime, NULL) > 0)
			continue;
		oldest_wstime = wstime;
		oldest_slot = i;
	}
	//PARA_CRIT_LOG("oldest slot: %d\n", oldest_slot);
	return oldest_slot;
}

__malloc static char *decoder_flags(void)
{
	int i;
	char flags[MAX_STREAM_SLOTS + 1];

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		char flag = '0';
		if (s->receiver_node)
			flag += 1;
		if (s->fns)
			flag += 2;
		if (s->wns)
			flag += 4;
		flags[i] = flag;
	}
	flags[MAX_STREAM_SLOTS] = '\0';
	return para_strdup(flags);
}

static int dump_commands(int fd)
{
	char *buf = para_strdup(""), *tmp = NULL;
	int i;
	ssize_t ret;

	FOR_EACH_COMMAND(i) {
		tmp = make_message("%s%s\t%s\n", buf, audiod_cmds[i].name,
			audiod_cmds[i].description);
		free(buf);
		buf = tmp;
	}
	ret = client_write(fd, buf);
	free(buf);
	return ret;
}

/*
 * command handlers don't close their fd on errors (ret < 0) so that
 * its caller can send an error message. Otherwise (ret >= 0) it's up
 * to each individual command to close the fd if necessary.
 */

static int com_help(int fd, int argc, char **argv)
{
	int i, ret;
	char *buf;
	const char *dflt = "No such command. Available commands:\n";

	if (argc < 2) {
		ret = dump_commands(fd);
		goto out;
	}
	FOR_EACH_COMMAND(i) {
		if (strcmp(audiod_cmds[i].name, argv[1]))
			continue;
		buf = make_message(
			"NAME\n\t%s -- %s\n"
			"SYNOPSIS\n\tpara_audioc %s\n"
			"DESCRIPTION\n%s\n",
			argv[1],
			audiod_cmds[i].description,
			audiod_cmds[i].usage,
			audiod_cmds[i].help
		);
		ret = client_write(fd, buf);
		free(buf);
		goto out;
	}
	ret = client_write(fd, dflt);
	if (ret > 0)
		ret = dump_commands(fd);
out:
	if (ret >= 0)
		close(fd);
	return ret;
}

static int com_tasks(int fd, __a_unused int argc, __a_unused char **argv)
{
	char *tl = get_task_list(&sched);
	int ret = 1;
	if (tl)
		ret = client_write(fd, tl);
	free(tl);
	if (ret > 0)
		close(fd);
	return ret;
}

static int com_stat(int fd, int argc, char **argv)
{
	int i, ret, parser_friendly = 0;
	uint64_t mask = 0;
	const uint64_t one = 1;
	struct para_buffer b = {.flags = 0};

	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		return ret;
	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strncmp(arg, "-p", 2)) {
			parser_friendly = 1;
			b.flags = PBF_SIZE_PREFIX;
			continue;
		}
	}
	if (i >= argc)
		mask--; /* set all bits */
	for (; i < argc; i++) {
		ret = stat_item_valid(argv[i]);
		if (ret < 0)
			return ret;
		mask |= (one << ret);
	}
	PARA_INFO_LOG("mask: 0x%llx\n", (long long unsigned)mask);
	FOR_EACH_STATUS_ITEM(i) {
		char *item = stat_item_values[i];
		if (!((one << i) & mask))
			continue;
		(void)WRITE_STATUS_ITEM(&b, i, "%s\n", item? item : "");
	}
	ret = client_write(fd, b.buf);
	if (ret >= 0)
		ret = stat_client_add(fd, mask, parser_friendly);
	free(b.buf);
	return ret;
}

static int com_grab(int fd, int argc, char **argv)
{
	return grab_client_new(fd, argc, argv, &sched);
}

__noreturn static int com_term(int fd, __a_unused int argc, __a_unused char **argv)
{
	close(fd);
	clean_exit(EXIT_SUCCESS, "terminating on user request");
}

static int com_on(int fd, __a_unused int argc, __a_unused char **argv)
{
	audiod_status = AUDIOD_ON;
	close(fd);
	return 1;
}

static int com_off(int fd, __a_unused int argc, __a_unused char **argv)
{
	audiod_status = AUDIOD_OFF;
	close(fd);
	return 1;
}

static int com_sb(int fd, __a_unused int argc, __a_unused char **argv)
{
	audiod_status = AUDIOD_STANDBY;
	close(fd);
	return 1;
}

static int com_cycle(int fd, int argc, char **argv)
{
	switch (audiod_status) {
		case  AUDIOD_ON:
			return com_sb(fd, argc, argv);
			break;
		case  AUDIOD_OFF:
			return com_on(fd, argc, argv);
			break;
		case  AUDIOD_STANDBY:
			return com_off(fd, argc, argv);
			break;
	}
	close(fd);
	return 1;
}

static int check_perms(uid_t uid)
{
	int i;

	if (!conf.user_allow_given)
		return 1;
	for (i = 0; i < conf.user_allow_given; i++)
		if (uid == conf.user_allow_arg[i])
			return 1;
	return -E_UCRED_PERM;
}

/**
 * Handle arriving connections on the local socket.
 *
 * \param accept_fd The fd to accept connections on.
 * \param rfds If \a accept_fd is not set in \a rfds, do nothing.
 *
 * This is called in each iteration of the select loop. If there is an incoming
 * connection on \a accept_fd, this function reads the command sent by the peer,
 * checks the connecting user's permissions by using unix socket credentials
 * (if supported by the OS) and calls the corresponding command handler if
 * permissions are OK.
 *
 * \return Positive on success, negative on errors, zero if there was no
 * connection to accept.
 *
 * \sa para_accept(), recv_cred_buffer()
 * */
int handle_connect(int accept_fd, fd_set *rfds)
{
	int i, argc, ret, clifd;
	char buf[MAXLINE], **argv = NULL;
	struct sockaddr_un unix_addr;
	uid_t uid;

	ret = para_accept(accept_fd, rfds, &unix_addr, sizeof(struct sockaddr_un), &clifd);
	if (ret <= 0)
		return ret;
	ret = recv_cred_buffer(clifd, buf, sizeof(buf) - 1);
	if (ret < 0)
		goto out;
	uid = ret;
	PARA_INFO_LOG("connection from user %i, buf: %s\n",  ret, buf);
	ret = check_perms(uid);
	if (ret < 0)
		goto out;
	ret = create_argv(buf, "\n", &argv);
	if (ret < 0)
		goto out;
	argc = ret;
	//PARA_INFO_LOG("argv[0]: %s, argc = %d\n", argv[0], argc);
	FOR_EACH_COMMAND(i) {
		if (strcmp(audiod_cmds[i].name, argv[0]))
			continue;
		ret = audiod_cmds[i].handler(clifd, argc, argv);
		goto out;
	}
	ret = -E_INVALID_AUDIOD_CMD;
out:
	free_argv(argv);
	if (clifd > 0 && ret < 0 && ret != -E_CLIENT_WRITE) {
		char *tmp = make_message("%s\n", para_strerror(-ret));
		client_write(clifd, tmp);
		free(tmp);
		close(clifd);
	}
	return ret;
}

/**
 * Send the current audiod status to all connected stat clients.
 */
void audiod_status_dump(void)
{
	int slot_num = get_play_time_slot_num();
	char *old, *new;

	old = stat_item_values[SI_PLAY_TIME];
	new = get_time_string(slot_num);
	if (new) {
		if (!old || strcmp(old, new)) {
			free(old);
			stat_item_values[SI_PLAY_TIME] = new;
			stat_client_write_item(SI_PLAY_TIME);
		} else
			free(new);
	}

	new = get_server_uptime_str(now);
	old = stat_item_values[SI_AUDIOD_UPTIME];
	if (!old || strcmp(old, new)) {
		free(old);
		stat_item_values[SI_AUDIOD_UPTIME] = new;
		stat_client_write_item(SI_AUDIOD_UPTIME);
	} else
		free(new);

	old = stat_item_values[SI_AUDIOD_STATUS];
	new = audiod_status_string();
	if (!old || strcmp(old, new)) {
		free(old);
		stat_item_values[SI_AUDIOD_STATUS] = new;
		stat_client_write_item(SI_AUDIOD_STATUS);
	} else
		free(new);

	old = stat_item_values[SI_DECODER_FLAGS];
	new = decoder_flags();
	if (!old || strcmp(old, new)) {
		free(old);
		stat_item_values[SI_DECODER_FLAGS] = new;
		stat_client_write_item(SI_DECODER_FLAGS);
	} else
		free(new);
}

/**
 * Flush and send all status items.
 *
 * Send to  each connected client the full status item list
 * with empty values.
 */
void clear_and_dump_items(void)
{
	int i;

	FOR_EACH_STATUS_ITEM(i) {
		free(stat_item_values[i]);
		stat_item_values[i] = NULL;
		stat_client_write_item(i);
	}
}
